#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import binascii
import json
import logging
import os
import time
from typing import Dict, Optional, Tuple

from redis import Redis

import skale.config as config
from skale.utils.web3_utils import get_receipt, MAX_WAITING_TIME
from skale.wallets import BaseWallet

logger = logging.getLogger(__name__)


class RedisAdapterError(Exception):
    pass


class DroppedError(RedisAdapterError):
    pass


class EmptyStatusError(RedisAdapterError):
    pass


class AdapterSendError(RedisAdapterError):
    pass


class AdapterWaitError(RedisAdapterError):
    pass


class RedisWalletAdapter(BaseWallet):
    ID_SIZE = 16

    def __init__(
        self,
        rs: Redis,
        pool: str,
        base_wallet: BaseWallet,
    ) -> None:
        self.rs = rs
        self.pool = pool
        self.wallet = base_wallet

    def sign(self, tx: Dict) -> Dict:
        return self.wallet.sign(tx)

    def sign_hash(self, unsigned_hash: str) -> str:
        return self.wallet.sign_hash(unsigned_hash)

    @property
    def address(self) -> str:
        return self.wallet.address

    @property
    def public_key(self) -> str:
        return self.wallet.public_key

    @classmethod
    def _make_raw_id(cls) -> bytes:
        prefix = b'tx-'
        unique = binascii.b2a_hex(os.urandom(cls.ID_SIZE // 2))
        return prefix + unique

    @classmethod
    def _make_score(cls, priority: int) -> int:
        ts = int(time.time())
        return priority * 10 ** len(str(ts)) + ts

    @classmethod
    def _make_record(
        cls,
        tx: Dict,
        score: int,
        multiplier: int = config.DEFAULT_GAS_MULTIPLIER
    ) -> Tuple[bytes, bytes]:
        tx_id = cls._make_raw_id()
        params = {
            'status': 'PROPOSED',
            'score': score,
            'multiplier': multiplier,
            'tx_hash': None,
            **tx
        }
        # Ensure gas will be restimated in TM
        params['gas'] = None
        record = json.dumps(params).encode('utf-8')
        return tx_id, record

    @classmethod
    def _to_raw_id(cls, tx_id: str) -> bytes:
        return tx_id.encode('utf-8')

    def _to_id(cls, raw_id: str) -> str:
        return raw_id.decode('utf-8')

    def sign_and_send(
        self,
        tx: Dict,
        multiplier: Optional[float] = None,
        priority: Optional[int] = None
    ) -> str:
        priority = priority or config.DEFAULT_PRIORITY
        try:
            logger.info(f'Sending {tx} to redis pool ...')
            score = self._make_score(priority)
            raw_id, tx_record = self._make_record(
                tx,
                score,
                multiplier=multiplier
            )
            pipe = self.rs.pipeline()
            logger.info(f'Adding tx {raw_id} to the pool')
            pipe.zadd(self.pool, {raw_id: score})
            logger.info(f'Saving tx {raw_id} record: {tx_record}')
            pipe.set(raw_id, tx_record)
            pipe.expire(raw_id, tx_record)
            pipe.execute()
            return self._to_id(raw_id)
        except Exception as err:
            logger.exception(f'Sending {tx} with redis wallet errored')
            raise AdapterSendError(err)

    def get_status(self, tx_id: str) -> str:
        return self.get_record(tx_id)['status']

    def get_record(self, tx_id: str) -> Dict:
        rid = self._to_raw_id(tx_id)
        return json.loads(self.rs.get(rid).decode('utf-8'))

    def wait(
        self,
        tx_id: str,
        blocks_to_wait: Optional[int] = None,
        timeout: int = MAX_WAITING_TIME
    ) -> Dict:
        start_ts = time.time()
        status = None

        while time.time() - start_ts < timeout:
            try:
                status = self.get_status(tx_id)
                if status == 'DROPPED':
                    break
                if status in ('SUCCESS', 'FAILED'):
                    r = self.get_record(tx_id)
                    return get_receipt(self.wallet._web3, r['tx_hash'])
            except Exception as err:
                logger.exception(f'Waiting for tx {tx_id} errored')
                raise AdapterWaitError(err)

        if status is None:
            raise EmptyStatusError('Tx status is None')
        if status == 'DROPPED':
            raise DroppedError('Tx was dropped after max retries')
        else:
            raise AdapterWaitError(f'Tx finished with status {status}')



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import wraps

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes


def formatter(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        res = method(self, *args, **kwargs)
        return {
            'earned': res[0],
            'end_month': res[1],
        }
    return wrapper


class Distributor(BaseContract):
    """Wrapper for Distributor.sol functions"""

    @formatter
    def get_earned_bounty_amount(self, validator_id: int, address: str) -> dict:
        """Get earned bounty amount for the validator

        :param validator_id: ID of the validator
        :type validator_id: int
        :returns: Earned bounty amount and end month
        :rtype: dict
        """
        return self.contract.functions.getAndUpdateEarnedBountyAmount(validator_id).call({
            'from': address
        })

    @formatter
    def get_earned_fee_amount(self, address: str) -> dict:
        """Get earned fee amount for the address

        :param address: Address of the validator
        :type address: str
        :returns: Earned bounty amount and end month
        :rtype: dict
        """
        return self.contract.functions.getEarnedFeeAmount().call({
            'from': address
        })

    @transaction_method
    def withdraw_bounty(self, validator_id: int, to: str) -> TxRes:
        """Withdraw earned bounty to specified address

        :param validator_id: ID of the validator
        :type validator_id: int
        :param to: Address to transfer bounty
        :type to: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.withdrawBounty(validator_id, to)

    @transaction_method
    def withdraw_fee(self, to: str) -> TxRes:
        """Withdraw earned fee to specified address

        :param to: Address to transfer bounty
        :type to: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.withdrawFee(to)



import pytest

from skale.schain_config.ports_allocation import (
    get_schain_base_port_on_node,
    calc_schain_base_port,
    get_schain_index_in_node
)
from skale.utils.exceptions import SChainNotFoundException
from skale.schain_config import PORTS_PER_SCHAIN
from tests.constants import (
    DEFAULT_NODE_NAME,
    DEFAULT_NODE_PORT,
    DEFAULT_SCHAIN_INDEX
)


def test_get_schain_base_port_on_node(skale, schain):
    schain_name = schain
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    schains_on_node = skale.schains.get_schains_for_node(node_id)
    schain_port_on_node = get_schain_base_port_on_node(
        schains_on_node,
        schain_name,
        DEFAULT_NODE_PORT
    )
    assert schain_port_on_node == DEFAULT_NODE_PORT


def test_get_schain_index_in_node(skale, schain):
    schain_name = schain
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    schains_on_node = skale.schains.get_schains_for_node(node_id)
    index = get_schain_index_in_node(schain_name, schains_on_node)
    assert isinstance(index, int)
    with pytest.raises(SChainNotFoundException):
        get_schain_index_in_node('ABCabc', schains_on_node)


def test_calc_schain_base_port():
    schain_base_port = calc_schain_base_port(DEFAULT_NODE_PORT, DEFAULT_SCHAIN_INDEX)
    schain_base_port_next = calc_schain_base_port(DEFAULT_NODE_PORT, DEFAULT_SCHAIN_INDEX + 1)
    schain_base_port_calc = schain_base_port + ((DEFAULT_SCHAIN_INDEX + 1) * PORTS_PER_SCHAIN)

    assert schain_base_port == DEFAULT_NODE_PORT
    assert schain_base_port_calc == schain_base_port_next



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" NodeRotation.sol functions """

import logging
import functools
import warnings
from dataclasses import dataclass

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes
from web3.exceptions import ContractLogicError


logger = logging.getLogger(__name__)


NO_PREVIOUS_NODE_EXCEPTION_TEXT = 'No previous node'


@dataclass
class Rotation:
    leaving_node_id: int
    new_node_id: int
    freeze_until: int
    rotation_counter: int


class NodeRotation(BaseContract):
    """Wrapper for NodeRotation.sol functions"""

    @property
    @functools.lru_cache()
    def schains(self):
        return self.skale.schains

    def get_rotation_obj(self, schain_name) -> Rotation:
        schain_id = self.schains.name_to_id(schain_name)
        rotation_data = self.contract.functions.getRotation(schain_id).call()
        return Rotation(*rotation_data)

    def get_rotation(self, schain_name):
        warnings.warn('Deprecated, will be removed in v6', DeprecationWarning)
        schain_id = self.schains.name_to_id(schain_name)
        rotation_data = self.contract.functions.getRotation(schain_id).call()
        return {
            'leaving_node': rotation_data[0],
            'new_node': rotation_data[1],
            'freeze_until': rotation_data[2],
            'rotation_id': rotation_data[3]
        }

    def get_leaving_history(self, node_id):
        raw_history = self.contract.functions.getLeavingHistory(node_id).call()
        history = [
            {
                'schain_id': schain[0],
                'finished_rotation': schain[1]
            }
            for schain in raw_history
        ]
        return history

    def get_schain_finish_ts(self, node_id: int, schain_name: str) -> int:
        raw_history = self.contract.functions.getLeavingHistory(node_id).call()
        schain_id = self.skale.schains.name_to_id(schain_name)
        finish_ts = next(
            (schain[1] for schain in raw_history if '0x' + schain[0].hex() == schain_id), None)
        if not finish_ts:
            return None
        exception = self.skale.schains_internal.check_exception(schain_name, node_id)
        if exception:
            rotation_delay = self.skale.constants_holder.get_rotation_delay()
            logger.info(f'Node {node_id} in exceptions array for {schain_name}, \
adding {rotation_delay} to {finish_ts}.')
            finish_ts += rotation_delay
        return finish_ts

    def is_rotation_in_progress(self, schain_name) -> bool:
        schain_id = self.schains.name_to_id(schain_name)
        return self.contract.functions.isRotationInProgress(schain_id).call()

    def is_new_node_found(self, schain_name) -> bool:
        schain_id = self.schains.name_to_id(schain_name)
        return self.contract.functions.isNewNodeFound(schain_id).call()

    def is_rotation_active(self, schain_name) -> bool:
        """
        The public function that tells whether rotation is in the active phase - the new group is
        already generated
        """
        finish_ts_reached = self.is_finish_ts_reached(schain_name)
        return self.is_rotation_in_progress(schain_name) and not finish_ts_reached

    def is_finish_ts_reached(self, schain_name) -> bool:
        rotation = self.skale.node_rotation.get_rotation_obj(schain_name)
        schain_finish_ts = self.get_schain_finish_ts(rotation.leaving_node_id, schain_name)

        if not schain_finish_ts:
            schain_finish_ts = 0

        latest_block = self.skale.web3.eth.getBlock('latest')
        current_ts = latest_block['timestamp']

        logger.info(f'current_ts: {current_ts}, schain_finish_ts: {schain_finish_ts}')
        return current_ts > schain_finish_ts

    def wait_for_new_node(self, schain_name):
        schain_id = self.schains.name_to_id(schain_name)
        return self.contract.functions.waitForNewNode(schain_id).call()

    @transaction_method
    def grant_role(self, role: bytes, owner: str) -> TxRes:
        return self.contract.functions.grantRole(role, owner)

    def has_role(self, role: bytes, address: str) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    def debugger_role(self):
        return self.contract.functions.DEBUGGER_ROLE().call()

    def get_previous_node(self, schain_name: str, node_id: int) -> int:
        schain_id = self.schains.name_to_id(schain_name)
        try:
            return self.contract.functions.getPreviousNode(schain_id, node_id).call()
        except ContractLogicError as e:
            if NO_PREVIOUS_NODE_EXCEPTION_TEXT in str(e):
                return None
            raise e



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE helper utilities """

import ipaddress
import json
import logging
import random
import socket
import string
import sys
from logging import Formatter, StreamHandler
from random import randint

from skale.config import ENV


logger = logging.getLogger(__name__)


def decapitalize(s):
    return s[:1].lower() + s[1:] if s else ''


def format_fields(fields, flist=False):
    """
        Transform array to object with passed fields
        Usage:
        @format(['field_name1', 'field_name2'])
        def my_method()
            return [0, 'Test']

        => {'field_name1': 0, 'field_name2': 'Test'}
    """

    def real_decorator(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)

            if result is None:
                return None

            if not isinstance(result, list) and not isinstance(result, tuple):
                return result

            if flist:
                formatted_list = []
                for item in result:
                    obj = {}
                    for i, field in enumerate(fields):
                        obj[field] = item[i]
                    formatted_list.append(obj)
                return formatted_list

            obj = {}
            for i, field in enumerate(fields):
                obj[field] = result[i]
            return obj

        return wrapper

    return real_decorator


def ip_from_bytes(packed: bytes) -> str:
    return socket.inet_ntoa(packed)


def ip_to_bytes(ip: str) -> bytes:  # pragma: no cover
    return socket.inet_aton(ip)


def is_valid_ipv4_address(address):
    try:
        ipaddress.IPv4Address(address)
    except ValueError:
        return False
    return True


def get_abi(abi_filepath=None):
    with open(abi_filepath) as data_file:
        return json.load(data_file)


def generate_nonce():  # pragma: no cover
    return randint(0, 65534)


def random_string(size=6, chars=string.ascii_lowercase):  # pragma: no cover
    return ''.join(random.choice(chars) for x in range(size))


def generate_random_ip():  # pragma: no cover
    return '.'.join('%s' % random.randint(0, 255) for i in range(4))


def generate_random_name(len=8):  # pragma: no cover
    return ''.join(
        random.choices(string.ascii_uppercase + string.digits, k=len))


def generate_random_port():  # pragma: no cover
    return random.randint(0, 60000)


def generate_custom_config(ip, ws_port):
    if not ip or not ws_port:
        raise ValueError(
            f'For custom init you should provide ip and ws_port: {ip}, {ws_port}'
        )
    return {
        'ip': ip,
        'ws_port': ws_port,
    }


def add_0x_prefix(bytes_string):  # pragma: no cover
    return '0x' + bytes_string


def rm_0x_prefix(bytes_string):
    if bytes_string.startswith('0x'):
        return bytes_string[2:]
    return bytes_string


def init_default_logger():  # pragma: no cover
    handlers = []
    formatter = Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    stream_handler = StreamHandler(sys.stderr)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    handlers.append(stream_handler)

    logging.basicConfig(level=logging.DEBUG, handlers=handlers)


def chunk(in_string, num_chunks):  # pragma: no cover
    chunk_size = len(in_string) // num_chunks
    if len(in_string) % num_chunks:
        chunk_size += 1
    iterator = iter(in_string)
    for _ in range(num_chunks):
        accumulator = []
        for _ in range(chunk_size):
            try:
                accumulator.append(next(iterator))
            except StopIteration:
                break
        yield ''.join(accumulator)


def split_public_key(public_key: str) -> list:
    public_key = rm_0x_prefix(public_key)
    pk_parts = list(chunk(public_key, 2))
    return list(map(bytes.fromhex, pk_parts))


def get_contracts_info(contracts_data):
    contracts_info = {}
    for contract_info in contracts_data:
        contracts_info[contract_info.name] = contract_info
    return contracts_info


def to_camel_case(snake_str):
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


def is_test_env():
    return "pytest" in sys.modules or ENV == 'test'



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes


class TokenState(BaseContract):
    """Wrapper for TokenState.sol functions"""

    def get_and_update_locked_amount(self, holder_address: str) -> int:
        """This method is for check quantity of `freezed` tokens
           :param holder_address: Address of the holder
           :type holder_address: str
           :returns:
           :rtype: int
        """
        return self.contract.functions.getAndUpdateLockedAmount(holder_address).call()

    @transaction_method
    def grant_role(self, role: bytes, owner: str) -> TxRes:
        return self.contract.functions.grantRole(role, owner)

    def has_role(self, role: bytes, address: str) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    def locker_manager_role(self):
        return self.contract.functions.LOCKER_MANAGER_ROLE().call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict
from eth_keys import keys
from web3 import Web3
from eth_account import messages

import skale.config as config
from skale.utils.web3_utils import get_eth_nonce, wait_for_receipt_by_blocks
from skale.wallets.common import BaseWallet, ensure_chain_id


def private_key_to_public(pr):
    pr_bytes = Web3.toBytes(hexstr=pr)
    pk = keys.PrivateKey(pr_bytes)
    return pk.public_key


def public_key_to_address(pk):
    hash = Web3.keccak(hexstr=str(pk))
    return to_checksum_address(Web3.toHex(hash[-20:]))


def private_key_to_address(pr):
    pk = private_key_to_public(pr)
    return public_key_to_address(pk)


def to_checksum_address(address):
    return Web3.toChecksumAddress(address)


def generate_wallet(web3):
    account = web3.eth.account.create()
    private_key = account.key.hex()
    return Web3Wallet(private_key, web3)


class Web3Wallet(BaseWallet):
    def __init__(self, private_key, web3):
        self._private_key = private_key
        self._public_key = private_key_to_public(self._private_key)
        self._address = public_key_to_address(self._public_key)

        self._web3 = web3

    def sign(self, tx_dict):
        if not tx_dict.get('nonce'):
            tx_dict['nonce'] = get_eth_nonce(self._web3, self._address)
        ensure_chain_id(tx_dict, self._web3)
        return self._web3.eth.account.sign_transaction(
            tx_dict,
            private_key=self._private_key
        )

    def sign_hash(self, unsigned_hash: str):
        unsigned_message = messages.encode_defunct(hexstr=unsigned_hash)
        return self._web3.eth.account.sign_message(
            unsigned_message,
            private_key=self._private_key
        )

    def sign_and_send(
        self,
        tx_dict: Dict,
        multiplier: int = config.DEFAULT_GAS_MULTIPLIER,
        priority: int = config.DEFAULT_PRIORITY
    ) -> str:
        signed_tx = self.sign(tx_dict)
        return self._web3.eth.sendRawTransaction(
            signed_tx.rawTransaction
        ).hex()

    @property
    def address(self):
        return self._address

    @property
    def public_key(self):
        return str(self._public_key)

    def wait(self, tx_hash: str, blocks_to_wait=None, timeout=None):
        return wait_for_receipt_by_blocks(
            self._web3,
            tx_hash,
            blocks_to_wait=blocks_to_wait,
            timeout=timeout
        )



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.allocator.allocator import TimeUnit

# manager test constants

DEFAULT_NODE_NAME = 'test_node'
SECOND_NODE_NAME = 'test_node_2'

DEFAULT_SCHAIN_NAME = 'test_schain_1'

DEFAULT_DOMAIN_NAME = 'skale.test'

INITIAL_DELEGATION_PERIOD = 2
D_VALIDATOR_ID = 1
D_VALIDATOR_NAME = 'test'
D_VALIDATOR_DESC = 'test'
D_VALIDATOR_FEE = 10
D_VALIDATOR_MIN_DEL = 1000000

D_DELEGATION_PERIOD = 2
D_STAKE_MULTIPLIER = 100
D_DELEGATION_INFO = 'test'

MONTH_IN_SECONDS = (60 * 60 * 24 * 31) + 100
FIRST_DELEGATION_MONTH = 1

# allocator test constants

TEST_VESTING_CLIFF = 6
TEST_TOTAL_VESTING_DURATION = 36
TEST_VESTING_INTERVAL_TIME_UNIT = TimeUnit.MONTH
TEST_VESTING_INTERVAL = 6
TEST_CAN_DELEGATE = True
TEST_IS_TERMINATABLE = True

TEST_START_MONTH = 8

WEI_MULTIPLIER = 10 ** 18

TEST_FULL_AMOUNT = 5000 * WEI_MULTIPLIER
TEST_LOCKUP_AMOUNT = 1000 * WEI_MULTIPLIER

POLL_INTERVAL = 2

TEST_SKALE_AMOUNT = 100000

D_PLAN_ID = 1



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
import time
from functools import partial, wraps
from typing import Dict, Optional

from web3 import Web3
from web3._utils.transactions import get_block_gas_limit

import skale.config as config
from skale.transactions.exceptions import TransactionError
from skale.transactions.result import TxRes
from skale.utils.web3_utils import get_eth_nonce
from skale.wallets.redis_wallet import RedisAdapterError


logger = logging.getLogger(__name__)


DEFAULT_ETH_SEND_GAS_LIMIT = 22000


def make_dry_run_call(skale, method, gas_limit=None, value=0) -> dict:
    opts = {
        'from': skale.wallet.address,
        'value': value
    }
    logger.info(
        f'Dry run tx: {method.fn_name}, '
        f'sender: {skale.wallet.address}, '
        f'wallet: {skale.wallet.__class__.__name__}, '
        f'value: {value}, '
    )

    try:
        if gas_limit:
            estimated_gas = gas_limit
            opts.update({'gas': gas_limit})
            method.call(opts)
        else:
            estimated_gas = estimate_gas(skale.web3, method, opts)
        logger.info(f'Estimated gas for {method.fn_name}: {estimated_gas}')
    except Exception as err:
        logger.error('Dry run for method failed with error', exc_info=err)
        return {'status': 0, 'error': str(err)}

    return {'status': 1, 'payload': estimated_gas}


def estimate_gas(web3, method, opts):
    try:
        block_gas_limit = get_block_gas_limit(web3)
    except AttributeError:
        block_gas_limit = get_block_gas_limit(web3)

    estimated_gas = method.estimateGas(
        opts,
        block_identifier='latest'
    )
    normalized_estimated_gas = int(
        estimated_gas * config.DEFAULT_GAS_MULTIPLIER
    )
    if normalized_estimated_gas > block_gas_limit:
        logger.warning(f'Estimate gas for {method.fn_name} - {normalized_estimated_gas} exceeds \
block gas limit, going to use block_gas_limit ({block_gas_limit}) for this transaction')
        return block_gas_limit
    return normalized_estimated_gas


def build_tx_dict(method, *args, **kwargs):
    base_fields = compose_base_fields(*args, **kwargs)
    return method.buildTransaction(base_fields)


def compose_base_fields(
    nonce: int,
    gas_limit: int,
    gas_price: Optional[int] = None,
    max_fee_per_gas: Optional[int] = None,
    max_priority_fee_per_gas: Optional[int] = None,
    value: Optional[int] = 0,
) -> Dict:
    fee_fields = {
        'gas': gas_limit,
        'nonce': nonce,
        'value': value
    }
    if max_priority_fee_per_gas is not None:
        fee_fields.update({
            'maxPriorityFeePerGas': max_priority_fee_per_gas,
            'maxFeePerGas': max_fee_per_gas
        })
        fee_fields.update({'type': 2})
    elif gas_price is not None:
        fee_fields.update({'gasPrice': gas_price})
        fee_fields.update({'type': 1})
    return fee_fields


def transaction_from_method(
    method,
    *,
    multiplier: Optional[float] = None,
    priority: Optional[int] = None,
    **kwargs
) -> str:
    tx = build_tx_dict(method, **kwargs)
    logger.info(
        f'Tx: {method.fn_name}, '
        f'Fields: {tx}, '
    )
    return tx


def compose_eth_transfer_tx(
    web3: Web3,
    from_address: str,
    to_address: str,
    value: int,
    **kwargs
) -> Dict:
    nonce = get_eth_nonce(web3, from_address)
    base_fields = compose_base_fields(
        nonce=nonce,
        gas_limit=DEFAULT_ETH_SEND_GAS_LIMIT,
        value=value,
        **kwargs
    )
    tx = {
        'from': from_address,
        'to': to_address,
        **base_fields
    }
    return tx


def retry_tx(tx=None, *, max_retries=3, timeout=-1):
    if tx is None:
        return partial(retry_tx, max_retries=3, timeout=timeout)

    @wraps(tx)
    def wrapper(*args, **kwargs):
        return run_tx_with_retry(
            tx, *args,
            max_retries=max_retries,
            retry_timeout=timeout, **kwargs
        )
    return wrapper


def run_tx_with_retry(transaction, *args, max_retries=3,
                      retry_timeout=-1,
                      **kwargs) -> TxRes:
    success = False
    attempt = 0
    tx_res = None
    exp_timeout = 1
    while not success and attempt < max_retries:
        try:
            tx_res = transaction(*args, **kwargs)
            tx_res.raise_for_status()
        except (RedisAdapterError, TransactionError) as err:
            logger.error(f'Tx attempt {attempt}/{max_retries} failed',
                         exc_info=err)
            timeout = exp_timeout if retry_timeout < 0 else exp_timeout
            time.sleep(timeout)
            exp_timeout *= 2
        else:
            success = True
        attempt += 1
    if success:
        logger.info(f'Tx {transaction.__name__} completed '
                    f'after {attempt}/{max_retries} retries')
    else:
        logger.error(
            f'Tx {transaction.__name__} failed after '
            f'{max_retries} retries')
    return tx_res



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2021-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes


class Wallets(BaseContract):
    def get_validator_balance(self, validator_id: int) -> int:
        """Returns SRW balance by validator id (in wei).

        :returns: SRW balance (wei)
        :rtype: int
        """
        return self.contract.functions.getValidatorBalance(validator_id).call()

    @transaction_method
    def recharge_validator_wallet(self, validator_id: int) -> TxRes:
        """Pass value kwarg (in wei) to the function when calling it"""
        return self.contract.functions.rechargeValidatorWallet(validator_id)

    @transaction_method
    def withdraw_funds_from_validator_wallet(self, amount: int) -> TxRes:
        return self.contract.functions.withdrawFundsFromValidatorWallet(amount)



from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes


class SlashingTable(BaseContract):
    """ Wrapper for SlashingTable.sol functions """

    @transaction_method
    def set_penalty(self, offense, penalty) -> TxRes:
        """ Set slashing penalty
        :param offense: reason of slashing
        :type offense: str
        :param penalty: penalty value to set
        :type penalty: int
        :rtype: TxRes
        """
        return self.contract.functions.setPenalty(offense, penalty)

    def get_penalty(self, offense) -> int:
        """ Get slashing penalty value
        :param offense: reason of slashing
        :type offense: str
        :rtype: int
        """
        return self.contract.functions.getPenalty(offense).call()

    @transaction_method
    def grant_role(self, role: bytes, address: str) -> TxRes:
        return self.contract.functions.grantRole(role, address)

    def penalty_setter_role(self) -> bytes:
        return self.contract.functions.PENALTY_SETTER_ROLE().call()

    def has_role(self, role: bytes, address: str) -> bool:
        return self.contract.functions.hasRole(role, address).call()
